import json
import decimal
import datetime


class LazyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, decimal.Decimal):
            return float(obj)
        elif isinstance(obj, datetime.datetime):
            return obj.strftime('%Y-%m-%d %H:%M:%S')
        elif isinstance(obj, datetime.date):
            return obj.strftime('%Y-%m-%d')
        super(LazyEncoder, self).default(obj)


def json2str(
        data: json,
        ensure_ascii: bool = False
):
    """
    在将json数据反序列化为str时，会遇到一些格式无法转换
    这里使用识别类型转换转为str
    目前支持类型：
    decimal --> str
    datetime.datetime --> str(%Y-%m-%d %H:%M:%S)
    datetime.date --> str(%Y-%m-%d)
    """
    return json.dumps(
        data,
        cls=LazyEncoder,
        ensure_ascii=ensure_ascii
    )



#!/usr/bin/env python3
# coding = utf8
"""
@ Author : ZeroSeeker
@ e-mail : zeroseeker@foxmail.com
@ GitHub : https://github.com/ZeroSeeker
@ Gitee : https://gitee.com/ZeroSeeker
"""
from ua_parser import user_agent_parser
from user_agents import parse
ua_match_ignore_values = ['Other', None]


def dict_values_match(
        dict_a: dict,
        dict_b: dict,
        ignore_values: list = None  # 忽略值，例如：['Other', None]
):
    """
    统计字典的value匹配量
    """
    dict_keys = list()
    dict_keys.extend(dict_a.keys())
    dict_keys.extend(dict_b.keys())
    dict_keys = set(dict_keys)
    count_all = 0  # 有效比对元素总数
    count_match = 0  # 对比结果匹配数
    for user_agent_key in dict_keys:
        each_value_a = dict_a.get(user_agent_key)
        each_value_b = dict_b.get(user_agent_key)
        if each_value_a == each_value_b:
            if ignore_values is not None and each_value_a in ignore_values:
                continue
            else:
                count_all += 1
                count_match += 1
        else:
            count_all += 1
    return {'count_all': count_all, 'count_match': count_match}


def ua_similarity(
        ua_a: str,
        ua_b: str,
        user_agent_match: bool = False,
        os_match: bool = True,
        device_match: bool = True
):
    """
    ua相似性算法，返回一个0-1之间的数，0为完全不匹配，1为完全匹配
    通过对user_agent、os、device中的元素进行比对，按照比对的成功个数/比对的总个数作为匹配率

    ua_match_ignore_values为无效或者无法失败的值的情况，不作为比对依据

    ua格式化后的格式为:
        {
            'user_agent': {
                'family': 'Other',
                'major': None,
                'minor': None,
                'patch': None
            },
            'os': {
                'family': 'Other',
                'major': None,
                'minor': None,
                'patch': None,
                'patch_minor': None
            },
            'device': {
                'family': 'Other',
                'brand': None,
                'model': None
            },
            'string': 'ua_string'
        }
    """
    if ua_a == ua_b:
        return 1
    else:
        ua_a_parser = user_agent_parser.Parse(ua_a)
        ua_b_parser = user_agent_parser.Parse(ua_b)

        match_res = list()
        if user_agent_match is True:
            user_agent_match_res = dict_values_match(
                dict_a=ua_a_parser.get('user_agent'),
                dict_b=ua_b_parser.get('user_agent'),
                ignore_values=ua_match_ignore_values
            )
            match_res.append(user_agent_match_res)
        else:
            pass
        if os_match is True:
            os_match_res = dict_values_match(
                dict_a=ua_a_parser.get('os'),
                dict_b=ua_b_parser.get('os'),
                ignore_values=ua_match_ignore_values
            )
            match_res.append(os_match_res)
        else:
            pass
        if device_match is True:
            device_match_res = dict_values_match(
                dict_a=ua_a_parser.get('device'),
                dict_b=ua_b_parser.get('device'),
                ignore_values=ua_match_ignore_values
            )
            match_res.append(device_match_res)
        else:
            pass

        # 计算匹配度
        count_all = 0  # 有效比对元素总数
        count_match = 0  # 对比结果匹配数
        for each_match_res in match_res:
            count_all += each_match_res.get('count_all')
            count_match += each_match_res.get('count_match')
        if count_all == 0:
            return 0
        else:
            return count_match/count_all


def ua_match(
        ua1: str,
        ua2: str,
        match_os: bool = True,
        match_device: bool = True
):
    """
    按照os和device匹配信息，匹配成功返回True，匹配失败返回False
    """
    user_agent1 = parse(ua1)
    user_agent2 = parse(ua2)
    user_agent1_os = user_agent1.os
    user_agent1_device = user_agent1.device
    user_agent2_os = user_agent2.os
    user_agent2_device = user_agent2.device
    if match_os and match_device:
        user_agent1_str = str(user_agent1_os) + str(user_agent1_device)
        user_agent2_str = str(user_agent2_os) + str(user_agent2_device)
    elif match_os and not match_device:
        user_agent1_str = str(user_agent1_os)
        user_agent2_str = str(user_agent2_os)
    elif not match_os and match_device:
        user_agent1_str = str(user_agent1_device)
        user_agent2_str = str(user_agent2_device)
    else:
        return False
    if user_agent1_str == user_agent2_str:
        return True
    else:
        return False



#!/usr/bin/env python3
# coding = utf8
"""
@ Author : ZeroSeeker
@ e-mail : zeroseeker@foxmail.com
@ GitHub : https://github.com/ZeroSeeker
@ Gitee : https://gitee.com/ZeroSeeker
"""


def make_proxies(
        port,
        host='localhost',
        protocol='http'
):
    """
    这里设置本地开启代理的时候需要将脚本开到对应端口的代理上，否则无法使用代理
    """
    proxy_meta = f"{protocol}://{host}:{port}"
    proxies = {
        "http": proxy_meta,
        "https": proxy_meta
    }
    return proxies



#!/usr/bin/env python3
# coding = utf8
"""
@ Author : ZeroSeeker
@ e-mail : zeroseeker@foxmail.com
@ GitHub : https://github.com/ZeroSeeker
@ Gitee : https://gitee.com/ZeroSeeker
"""
import copy
from collections import OrderedDict
from urllib import parse


def dict_list_ranker_x(dict_list, rank_by_list):
    """
    对dict组成的list按照某个元素重新排序
    可对多个元素排序,目前支持2个
    rank_by_list：[['a','asc']]
    :return:
    """
    ranker_res_list = list()
    for each_dict in dict_list:
        each_rank_by_value_1 = each_dict.get(rank_by_list[0][0])
        each_rank_by_value_2 = each_dict.get(rank_by_list[1][0])
        if len(ranker_res_list) == 0:
            ranker_res_list.append(each_dict)
        else:
            pos_num = 0
            for each_ranker_res in ranker_res_list:
                each_ranker_res_rank_by_value_1 = each_ranker_res.get(rank_by_list[0][0])
                each_ranker_res_rank_by_value_2 = each_ranker_res.get(rank_by_list[1][0])
                if rank_by_list[0][1].lower() == 'asc' and rank_by_list[1][1].lower() == 'desc':
                    if each_rank_by_value_1 > each_ranker_res_rank_by_value_1:
                        pos_num += 1
                    elif each_rank_by_value_1 == each_ranker_res_rank_by_value_1:
                        if each_rank_by_value_2 < each_ranker_res_rank_by_value_2:
                            pos_num += 1
                        else:
                            pass
                    elif each_rank_by_value_1 < each_ranker_res_rank_by_value_1:
                        pass
            ranker_res_list.insert(pos_num, each_dict)
    return ranker_res_list


def dict_list_ranker(dict_list, rank_by_list):
    """
    对dict组成的list按照某个元素重新排序
    可对多个元素排序,目前支持1个
    rank_by_list：[['a','asc']]
    :return:
    """
    ranker_res_list = list()
    for each_dict in dict_list:
        each_rank_by_value_1 = each_dict.get(rank_by_list[0][0])
        if len(ranker_res_list) == 0:  # 如果结果列表为空，放入初始值
            ranker_res_list.append(each_dict)
        else:
            pos_num = 0
            for each_ranker_res in ranker_res_list:
                each_ranker_res_rank_by_value_1 = each_ranker_res.get(rank_by_list[0][0])
                if each_rank_by_value_1 > each_ranker_res_rank_by_value_1:
                    pos_num += 1
                else:
                    pass
            ranker_res_list.insert(pos_num, each_dict)
    if rank_by_list[0][1].lower() == 'asc':
        return ranker_res_list
    else:
        ranker_res_list.reverse()
        return ranker_res_list


def dict_list_group(list_in, by):
    """
    将dict组成的list按照某个key的值分组，分组后组成新的dict嵌套list
    :param list_in:
    :param by:
    :return:
    """
    if not list_in:
        return
    else:
        by_list = list()
        for each_dict in list_in:
            by_value = each_dict.get(by)
            by_list.append(by_value)
        by_set = set(by_list)
        group_dict = dict()
        for each_by in by_set:
            group_dict[each_by] = list()
        for each_dict in list_in:
            by_value = each_dict.get(by)
            group_dict[by_value].extend([each_dict])
        return group_dict


def dict_key_f(in_dict):
    """
    将以驼峰方式命名的dict转换为下划线方式
    """
    out_dict = dict()
    for key, value in in_dict.items():
        # key_split = list(key)
        # print(key, key_split)
        key_split_new = list()
        for index, temp_key in enumerate(key):
            if temp_key.isupper() is True:
                if index == 0:
                    key_split_new.append(temp_key.lower())
                else:
                    key_split_new.append('_')
                    key_split_new.append(temp_key.lower())
            else:
                key_split_new.append(temp_key)
        key_new = ''.join(key_split_new)
        # print(key_new, key_split_new)
        out_dict[key_new] = value
    return out_dict


def list_dict_key_f(in_list):
    """
    将以驼峰方式命名的list嵌套dict转换为下划线方式
    """
    out_list = list()
    for each_dict in in_list:
        out_dict = dict_key_f(each_dict)
        out_list.append(out_dict)
    return out_list


def dict_value_parse(dict_in):
    # 对字典的值进行解码
    dict_out = dict()
    for key, value in dict_in.items():
        if isinstance(value, str):
            dict_out[key] = parse.unquote(value)
        else:
            dict_out[key] = value
    return dict_out


def list_dict_value_parse(list_in):
    # 在dict上面封装一层list，解码
    list_out = list()
    for each_list_in in list_in:
        dict_out = dict_value_parse(dict_in=each_list_in)
        list_out.append(dict_out)
    return list_out


def dict_tiler(dict_in, connector_str='-'):
    """
    字典展开器
    只展开一层
    将嵌套了一层的dict铺展开，键值之间用"-"连接
    针对纯dict形式
    :param dict_in: 输入一个待处理的dict
    :param connector_str: 输入连接符
    :return: 输出处理好的dict
    """
    dict_f = dict()
    if dict_in is None:
        return
    elif len(dict_in) == 0:
        return
    else:
        for key in dict_in:
            value = dict_in.get(key)
            if isinstance(value, dict):  # 如果某个key的值是dict，则展开这组
                for each_value_key in value:
                    dict_f['%s%s%s' % (key, connector_str, each_value_key)] = value.get(each_value_key)
            else:
                dict_f[key] = value
        return dict_f


def list_dict_tiler(list_in, connector_str='-'):
    list_out = list()
    for each in list_in:
        dict_f = dict_tiler(dict_in=each, connector_str=connector_str)
        list_out.append(dict_f)
    return list_out


def list_same_order_dict(
        list_data: list = None,
        na_value=None,
        keys_sort: list = None
) -> list:
    """
    list重排格式化，将单层dict组成的list按照全部key的顺序排序并补充缺失值
    :param list_data: 原始list
    :param na_value: 缺失值的默认值
    :param keys_sort: 指定顺序,按照指定列的顺序先排，不包含的向后排
    :return list_data_f: 处理后的结果

    test data:
        [
            {'a': 1},
            {'b': 2},
            {'b': 2, 'a': 1},
            {'c': 21}
        ]
    """
    keys = list()  # 存储所有的key
    for each_dict in list_data:
        # 遍历list中的每个dict
        each_keys = list(each_dict.keys())  # 提取当前dict的所有key
        for each_key in each_keys:
            # 遍历，得到全量唯一的key
            if each_key in keys:
                continue
            else:
                keys.append(each_key)
    keys.sort()  # 先按照升序排序一次
    if keys_sort:
        keys_new = copy.deepcopy(keys_sort)  # 先直接拷贝一下排序表
        for each_key in keys:
            # 遍历生成的key列表
            if each_key in keys_new:
                # 如果遍历的key已存在，将跳过
                continue
            else:
                # 如果遍历的key不已存在，则追加
                keys_new.append(each_key)
    else:
        keys_new = copy.deepcopy(keys)
    list_data_f = list()  # 存储处理后的数据
    for each_dict in list_data:
        each_dict_order = OrderedDict()  # 有序字典
        for each_key in keys_new:
            each_dict_order[each_key] = each_dict.get(each_key, na_value)
        list_data_f.append(each_dict_order)
    return list_data_f


def list_dict_filter(
        list_in: list,
        filter_key,
        filter_value
):
    """
    从[{},{}]中按照某个key-value条件筛选出符合条件的记录
    """
    list_out = list(filter(lambda x: x[filter_key] == filter_value, list_in))
    return list_out


def key_max_value(
        list_in: list,
        key
):
    """
    求list嵌套的dict中某个key的最大值
    """
    max_value = None
    for each in list_in:
        key_value = each.get(key)
        if key_value:
            if max_value is None or key_value > max_value:
                max_value = copy.deepcopy(key_value)
            else:
                continue
        else:
            continue
    return max_value


def key_min_value(
        list_in: list,
        key
):
    """
    求list嵌套的dict中某个key的最小值
    """
    min_value = None
    for each in list_in:
        key_value = each.get(key)
        if key_value:
            if min_value is None or key_value < min_value:
                min_value = copy.deepcopy(key_value)
            else:
                continue
        else:
            continue
    return min_value


def get_value_list(
        list_in: list,
        key,
        deepcopy: bool = True
):
    """
    提取list嵌套的字典中某个key的值列表
    """
    res = []
    if list_in:
        for each in list_in:
            if isinstance(each, dict):
                each_value  = each.get(key)
                if each_value:
                    res.append(each_value)
                else:
                    continue
            else:
                continue
        if deepcopy:
            return copy.deepcopy(res)
        else:
            return res
    else:
        return []




#!/usr/bin/env python3
# coding = utf8
"""
@ Author : ZeroSeeker
@ e-mail : zeroseeker@foxmail.com
@ GitHub : https://github.com/ZeroSeeker
@ Gitee : https://gitee.com/ZeroSeeker
"""
import string
import random


def random_str(
        str_length: int = 8,
        ascii_lowercase: bool = True,  # 小写字母
        ascii_uppercase: bool = True,  # 大写字母
        numbers: bool = True,  # 0-9的数字
        repeat: bool = True
):
    """
    按照规则生成某个长度的随机字符串
    :param str_length: 要生成的字符串长度
    :param ascii_lowercase: 包含小写字母
    :param ascii_uppercase: 包含大写字母
    :param numbers: 包含0-9的数字
    :param repeat: 是否重复，默认为True，False不重复采样，True可重复采样
    """
    random_sample = ""
    if ascii_lowercase is True:
        random_sample += string.ascii_lowercase
    if ascii_uppercase is True:
        random_sample += string.ascii_uppercase
    if numbers is True:
        random_sample += string.digits
    if repeat:
        # 可重复采样
        value = ''
        for _ in range(str_length):
            value += random.sample(
                random_sample,
                1
            )[0]
    else:
        # 不重复采样
        value = ''.join(
            random.sample(
                random_sample,
                str_length
            )
        )
    return value



import re


def find_between(
        text: str,
        left: str,
        right: str,
        until_line_end: bool = False  # 是否匹配到行尾
):
    """
    找两个之间
    """
    if until_line_end:
        return re.findall(pattern=f'{left}(.*?){right}$', string=text, flags=re.S)
    else:
        return re.findall(pattern=f'{left}(.*?){right}', string=text, flags=re.S)


def find_all(
        text: str,
        left: str,
        right: str,
        until_line_end: bool = False  # 是否匹配到行尾
):
    """
    找一个之后所有
    """
    if until_line_end:
        return re.findall(pattern=f'{left}(.*){right}$', string=text, flags=re.S)
    else:
        return re.findall(pattern=f'{left}(.*){right}', string=text, flags=re.S)


def find_chs(
        text: str
):
    """
    提取中文，目前不能去掉括号
    """
    return re.sub(pattern="[A-Za-z0-9\!\%\[\]\,\。\ \']", repl="", string=text)



#!/usr/bin/env python3
# coding = utf8
"""
@ Author : ZeroSeeker
@ e-mail : zeroseeker@foxmail.com
@ GitHub : https://github.com/ZeroSeeker
@ Gitee : https://gitee.com/ZeroSeeker
"""
import showlog
import copy
import re


def get_split(
        content: str,
        split_mark: list
):
    """
    将字符串按照指定的字符分割，返回一个list结果
    """
    split_res = list()
    for each_mark in split_mark:
        if len(split_res) == 0:
            split_res = content.split(each_mark)
        else:
            temp_split_res = list()
            for each_split_res in split_res:
                temp_split_res.extend(each_split_res.split(each_mark))
            split_res = copy.deepcopy(temp_split_res)
    return split_res


def get_url(
        content
):
    """
    从文本中获取url
    """
    content_split = get_split(content=content, split_mark=[' ', '\n'])
    url_list = list()
    for each_split in content_split:
        if 'http' in each_split and '//' in each_split:
            # 这是一个链接
            input_text_find = re.findall('http(.*?)$', each_split, re.S)
            if len(input_text_find) > 0:
                input_url = 'http%s' % input_text_find[0]
                url_list.append(input_url)
            else:
                showlog.warning('解析链接失败')
                continue
        else:
            # 这不是一个链接
            continue
    return url_list


def path_clean(content):
    """
    清除路径前后可能出现的引号
    """
    if not content:
        return content
    elif content[0] == '"' and content[-1] == '"':
        content = content[1:-1]
    elif content[0] == '“' and content[-1] == '”':
        content = content[1:-1]
    elif content[0] == "'" and content[-1] == "'":
        content = content[1:-1]
    else:
        pass
    return content


def file_name_format(text):
    """
    对文件名处理，避免特殊字符导致的无法存储的问题
    """
    text_new = re.sub(r'[\\/:*?"<>|\r\n]+', "_", text)  # 规避windows非法字符
    return text_new


def name_head_list(
        your_name: str
):
    """
    获取姓名的首字母缩写的所有可能列表
    :param your_name:
    :return:
    """
    from xpinyin import Pinyin
    p = Pinyin()
    name_pinyins = p.get_pinyins(your_name)
    name_heads = list()
    for each_name_pinyin in name_pinyins:
        name_pinyin_split = each_name_pinyin.split('-')
        name_head = ""
        for each_split in name_pinyin_split:
            each_split_0 = each_split[0]
            name_head += each_split_0
        name_heads.append(name_head)
    return name_heads



#!/usr/bin/env python3
# coding = utf8
"""
@ Author : ZeroSeeker
@ e-mail : zeroseeker@foxmail.com
@ GitHub : https://github.com/ZeroSeeker
@ Gitee : https://gitee.com/ZeroSeeker
"""
import json

import requests


def get_public_ip() -> str:
    """
    获取当前网络公网ip地址
    备用地址：http://www.3322.org/dyndns/getip
    """
    import requests
    import json
    origin_ip = ''
    try:
        request_url = "http://httpbin.org/ip"
        response = requests.get(url=request_url)
        origin_ip = json.loads(response.text).get("origin")
    finally:
        return origin_ip


def get_local_ip() -> str:
    """
    获取内网ip地址
    """
    import socket
    ip = ''
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(('8.8.8.8', 80))
        ip = s.getsockname()[0]
    finally:
        s.close()
        return ip


def get_ip() -> dict:
    """
    获取当前网络ip地址（含有公网ip和内网ip）
    """
    origin_ip = get_public_ip()  # 获取公网ip
    local_ip = get_local_ip()  # 获取内网ip
    return {'origin_ip': origin_ip, 'local_ip': local_ip}


def get_ip_addr(ip: str):
    """
    查询ip归属地
    """
    api_url = f'http://whois.pconline.com.cn/ipJson.jsp?ip={ip}&json=true'
    response = requests.get(api_url)
    response_text = response.text.replace("\\", "-")
    addr = json.loads(response_text)['addr']
    return addr



#!/usr/bin/env python3
# coding = utf8
"""
@ Author : ZeroSeeker
@ e-mail : zeroseeker@foxmail.com
@ GitHub : https://github.com/ZeroSeeker
@ Gitee : https://gitee.com/ZeroSeeker
"""
import platform
import os


def basic(
        file_name: str
):
    inner_file_name = file_name.lower()  # 不区分大小写

    if platform.system() == 'Windows':  # Windows
        basic_return = {
            'sys_support': True,
            'path_separator': '\\',
            'env_path': 'C:\\env\\',
            'file_dir': 'C:\\env\\%s' % inner_file_name
        }
        return basic_return
    elif platform.system() == 'Darwin':  # macOS
        basic_return = {
            'sys_support': True,
            'path_separator': '/',
            'env_path': '/Users/env/',
            'file_dir': '/Users/env/%s' % inner_file_name
        }
        return basic_return
    elif platform.system() == 'Linux':  # Linux
        basic_return = {
            'sys_support': True,
            'path_separator': '/',
            'env_path': '/env/',
            'file_dir': '/env/%s' % inner_file_name
        }
        return basic_return
    else:
        basic_return = {
            'sys_support': False,
            'path_separator': '',
            'env_path': '',
            'file_dir': ''
        }
        return basic_return


def read(
        file_name: str
):
    """
    file_name:环境文件名称，包括后缀名，不区分，例如：
        mysql.env
        mongo.env
        redis.env

    环境文件的内容是以航区分，以=符号指定，例如：HOST=192.168.0.1
    读取的结果是一个dict，将原来的行按照=符号组成键值对，例如：{"HOST": "192.168.0.1"}

    环境文件路径支持：
        Windows：
            C:\
        macOS:
            /Users/env/
        Linux:
            /env/
    """
    env_dict = dict()
    file_name_lower = file_name.lower()
    basic_info = basic(file_name=file_name)
    env_path = basic_info['env_path']
    env_file_list = os.listdir(env_path)
    for each_env_file in env_file_list:
        if file_name_lower == each_env_file.lower():
            env_file_dir = '%s%s' % (env_path, each_env_file)
            f = open(env_file_dir, encoding='utf-8')
            file_read = f.read()
            lines = file_read.split('\n')
            for each_line in lines:
                if '=' in each_line:
                    each_line_split = each_line.split(sep='=', maxsplit=1)  # 只拆分一次，防止有多个=影响
                    env_dict[each_line_split[0]] = each_line_split[1]
                else:
                    pass
            return env_dict
        else:
            continue


def get_default_env():
    """
    读取默认环境信息
    存在返回：{'ENV': 'DEV', 'MSG': '开发环境'}
    不存在返回：{'ENV': None, 'MSG': None}
    """
    default_env = read(file_name='DEFAULT_ENV.env')
    if default_env is None:
        return {'ENV': None, 'MSG': None}
    else:
        return default_env



from xml.etree import ElementTree
import re


def xml2dict(xml_str: str):
    """
    XML代码转dict
    在格式化之前，先抽取出xml标签内的内容，防止有干扰
    参考：https://docs.python.org/zh-cn/3.8/library/xml.etree.elementtree.html
    https://blog.csdn.net/qdPython/article/details/115520713
    :param xml_str:
    :return:
    """
    res_dict = dict()
    xml_str_res = re.findall(r'<xml>(.*?)</xml>', xml_str, re.S)
    if xml_str_res:
        xml_str_process = f"<xml>{xml_str_res[0]}</xml>"
        root = ElementTree.fromstring(xml_str_process)
        for child in root:
            # print(child.tag, child.attrib, child.text)
            res_dict[child.tag] = child.text
        return res_dict
    else:
        return res_dict



import random
import uuid
import time
import datetime
import os

uuid_node = random.randint(100000000, 999999999)  # 避免泄密设备信息


# def random_mac_uuid():
#     """
#     uuid1
#     由MAC地址、当前时间戳、随机数生成。可以保证全球范围内的唯一性，
#     但MAC的使用同时带来安全性问题，局域网中可以使用IP来代替MAC。
#     这里使用随机数代替
#     """
#     return str(uuid.uuid1(node=uuid_node))  # 替换原来的MAC地址为随机数


# def timestamp_uuid():
#     """
#     使用纳秒级别时间戳+uuid作为id
#     """
#     return f'{time.time_ns()}_{random_mac_uuid()}'


def make_id():
    """
    基于毫秒级时间和随机字符串拼接生成id，默认生成长度为50
    """
    return MakeId()


class MakeId:
    def __init__(self):
        utcnow = datetime.datetime.utcnow()  # UTC时间
        utcnow_str = utcnow.strftime('%Y%m%d%H%M%S%f')  # 时间格式为：[4位年，2位月，2位日，2位时，2位分，2位秒，6位毫秒]
        urandom_hex = os.urandom(15).hex()  # 算出来的字符长度是输入长度的2倍
        self.utcnow = utcnow
        self.utcnow_str = utcnow_str
        self.urandom_hex = urandom_hex
        self.id = f"{utcnow_str}{urandom_hex}"
        self.random_uuid = str(uuid.uuid1(node=uuid_node))
        self.mac_uuid = str(uuid.uuid1())





def decode(
        string,
        decode_list: list = None
):
    if not decode_list:
        decode_list = [
            "UTF-8",
            "ISO-8859-1",
            "gb18030",
            "gb2312",
            "gbk",
            "latin-1"
        ]
    for each_decode in decode_list:
        try:
            return string.decode(each_decode)
        except UnicodeDecodeError:
            pass
    pass






def split(
        list_in: list,
        split_gap: int
):
    """
    将输入的list按照split_gap进行分割
    """
    list_out = list()
    split_num = 0
    for i in range(len(list_in)):
        if i % split_gap == 0:
            list_out.append(list())
            split_num += 1
        list_out[split_num-1].append(list_in[i])
    return list_out



#!/usr/bin/env python3
# coding = utf8
"""
@ Author : ZeroSeeker
@ e-mail : zeroseeker@foxmail.com
@ GitHub : https://github.com/ZeroSeeker
@ Gitee : https://gitee.com/ZeroSeeker
"""
import copy

from . import lazypath
import collections
import subprocess
import requests
import platform
import datetime
import openpyxl
import showlog
import time
import xlrd
import json
import sys
import os
# import zipfile
from rich.progress import Progress
from requests import exceptions

headers_default = {"User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:68.0) Gecko/20100101 Firefox/68.0"}


def delete(
        file,
        postfix: str = None,
        path: str = None,
):
    """
    删除文件
    """
    if path:
        # 如果指定了路径，就加上路径
        file_dir = os.path.join(path, file)
    else:
        # 如果没指定路径，就直接使用文件名
        file_dir = file

    if postfix:
        # 如果指定了后缀名，就加上后缀名
        file_dir = f'{file}.{postfix}'
    else:
        # 如果没指定后缀名，就忽略
        pass
    os.remove(file_dir)


def get_suffix(
        file_name: str
):
    return file_name.split('.')[-1]  # 获取文件后缀名


def file_rename(
        before_name: str,  # 原名称
        after_name: str  # 重命名名称
):
    """
    重命名文件，注意是完整文件路径
    :param before_name:
    :param after_name:
    :return:
    """
    try:
        os.rename(before_name, after_name)
        return True
    except Exception as e:
        print(e)
        print('rename file fail\r\n')
        return False


def open_folder(
        folder: str = os.path.dirname(os.path.abspath(__file__))  # 默认为当前路径
):
    """
    功能：打开路径
    """
    if platform.system() == 'Windows':
        os.startfile(folder)  # Windows上打开文件
    else:
        subprocess.check_call(['open', folder])  # 非Windows上打开文件


def download(
        url,
        filename: str = None,
        suffix_name: str = None,
        headers: dict = None,
        path: str = "download",
        proxies=None,
        size_limit: int = None,
        range_start: int = None,
        range_end: int = None,
        overwrite: bool = False,
        verify: bool = True
):
    """
    实现文件下载功能，可指定url、文件名、后缀名、请求头、文件保存路径
    :param url:
    :param filename:文件名
    :param suffix_name:后缀名
    :param headers:请求头
    :param path:文件保存路径
    :param proxies:代理
    :param size_limit:尺寸限制
    :param range_start:开始位置
    :param range_end:结束位置
    :param overwrite: 覆盖
    :param verify: 验证证书
    :return:
    """
    if not headers:
        headers_local = headers_default
    else:
        headers_local = headers

    if not range_start and not range_end:
        range_start = 0
        range_info = None
    elif range_start is not None and range_end is None:
        range_info = 'bytes=%d-' % range_start  # 从这里向后
    elif range_start is None and range_end is not None:
        range_start = 0
        range_info = 'bytes=0-%d' % range_end
    else:
        range_info = 'bytes=%d-%d' % (range_start, range_end)

    if not range_info:
        pass
    else:
        headers_local['Range'] = range_info
    # 获取文件的基本信息
    response = requests.get(
        url=url,
        headers=headers_local,
        stream=True,
        proxies=proxies,
        verify=verify
    )
    total_length = response.headers.get('content-length', '0')  # 文件大小
    content_type = response.headers.get('content-type')  # 文件类型
    content_disposition = response.headers.get('content-disposition')  # 文件名及类型
    filename_default = 'unknown_' + str(time.time())
    if content_disposition is not None:
        content_dispositions = content_disposition.replace(' ', '').split(';')
        for each_content_disposition in content_dispositions:
            if 'filename' in each_content_disposition:
                each_content_disposition_split = each_content_disposition.split(sep='=', maxsplit=1)  # 只拆分一次，防止有多个=影响
                filename_default_full = each_content_disposition_split[1]
                filename_default = filename_default_full[:filename_default_full.rfind('.')]  # 解析文件名
                suffix_name = filename_default_full[filename_default_full.rfind('.')+1:]  # 解析文件后缀
            else:
                pass
    else:
        pass

    if suffix_name is None:
        # 尝试自动获取文件后缀名
        suffix_name = content_type.split('/')[1]

    if filename is None:
        download_file_name = str(filename_default) + "." + str(suffix_name)
    else:
        download_file_name = str(filename) + "." + str(suffix_name)

    if path is None:
        path_local = download_file_name
    else:
        if os.path.exists(path):
            pass
        else:
            os.makedirs(path)
        path_local = os.path.join(path, download_file_name)

    if not range_start:
        temp_size = 0  # 已经下载文件大小
    else:
        temp_size = range_start + 0  # 已经下载文件大小
    chunk_size = 1024 * 1024  # 分割文件大小，字节B
    total_size = int(total_length)  # 文件总大小
    # total_size_mb = round(total_size / (1024 * 1024), 2)  # 换算到MB的文件大小
    # 添加文件大小控制，跳过下载超大文件
    if size_limit is None:
        pass
    else:
        if total_size > size_limit:
            return
        else:
            pass

    is_finish = False

    if overwrite and os.path.exists(path_local):
        delete(file=path_local)
    elif not overwrite and os.path.exists(path_local):
        # 这里将对已存在的文件名重新命名
        rename_count = 1
        while True:
            download_file_name_rename = f'{download_file_name.split(".")[0]}({rename_count}).{download_file_name.split(".")[1]}'
            path_local_rename = os.path.join(path, download_file_name_rename)
            if os.path.exists(path_local_rename):
                rename_count += 1
                continue
            else:
                break
        path_local = copy.deepcopy(path_local_rename)
    else:
        pass

    with open(path_local, "ab") as f:  # wb新建文件，a追加
        with Progress() as progress:
            task = progress.add_task(description="[red]Downloading...", total=total_size)
            for chunk in response.iter_content(chunk_size=chunk_size):
                try:
                    if not chunk:
                        if temp_size >= total_size:
                            is_finish = True
                        else:
                            is_finish = False
                        break
                    else:
                        f.write(chunk)
                        f.flush()
                        progress.update(task, advance=chunk_size)
                        if temp_size >= total_size:
                            is_finish = True
                        else:
                            is_finish = False
                except:
                    showlog.error('')
    print("\n  ==> 文件已全部下载完成，保存位置:", path_local)
    res_dict = {
        'file_dir': path_local,
        'is_finish': is_finish,
        'temp_size': temp_size
    }
    return res_dict


def safe_download(
        url,
        filename=None,
        suffix_name=None,
        headers=None,
        path="download",
        proxies=None,
        size_limit=None,
        range_start=None,
        range_end=None,
        verify: bool = False
):
    while True:
        try:
            download_response = download(
                url=url,
                filename=filename,
                suffix_name=suffix_name,
                headers=headers,
                path=path,
                proxies=proxies,
                size_limit=size_limit,
                range_start=range_start,
                range_end=range_end,
                overwrite=False,
                verify=verify
            )
            if download_response.get('is_finish') is True:
                local_file_dir = download_response.get('file_dir')
                return local_file_dir
            else:
                print(':( 下载中断')
                range_start = download_response.get('temp_size')
                time.sleep(1)
                print('将继续下载（断点续传）...')
        except:
            print(':( 下载中断，将重新下载...')
            time.sleep(1)


def read(
        file,
        postfix: str = None,
        path: str = None,
        json_auto: bool = False,
        read_lines: bool = False
):
    """
    读取文件
    json_auto：json格式自动转换
    """
    if path:
        # 如果指定了路径，就加上路径
        file_dir = os.path.join(path, file)
    else:
        # 如果没指定路径，就直接使用文件名
        file_dir = file

    if postfix:
        # 如果指定了后缀名，就加上后缀名
        file_dir = f'{file}.{postfix}'
    else:
        # 如果没指定后缀名，就忽略
        pass

    if not os.path.exists(file_dir):
        return

    if read_lines:
        with open(file=file_dir, mode='r', encoding='utf-8') as f:
            content = f.readlines()
    else:
        f = open(file=file_dir, mode='r', encoding='utf-8')
        content = f.read()

    if content:
        if json_auto:
            if isinstance(content, str):
                return json.loads(content)
            else:
                json_content = list()
                for each_line in content:
                    json_content.append(json.loads(each_line))
                return json_content
        else:
            return content
    else:
        return content


def save(
        file,
        content,
        suffix: str = None,
        path: str = None,
        overwrite: bool = True,
        encoding: str = 'utf-8'
) -> str:
    """
    保存文件
    """
    if path:
        # 如果指定了路径，就加上路径
        lazypath.make_path(path)
        file_dir = os.path.join(path, file)
    else:
        # 如果没指定路径，就直接使用文件名
        file_dir = file

    if suffix:
        # 如果指定了后缀名，就加上后缀名
        file_dir = f'{file_dir}.{suffix}'
    else:
        # 如果没指定后缀名，就忽略
        pass

    if overwrite is True:
        write_mode = "w"  # 覆盖
    else:
        write_mode = "a"  # 追加

    f = open(
        file=file_dir,
        mode=write_mode,
        encoding=encoding
    )
    f.write(content)
    f.close()
    return file_dir


def read_(_source_file):
    # >>读取数据【方式：一次性全部读取】------------------------------------------------------------
    data = xlrd.open_workbook(_source_file)  # 打开表
    res = list()
    # table = data.sheets()[0]  # 默认使用第一张表格
    for table in data.sheets():
        nrows = table.nrows  # 获取行数
        ncols = table.ncols  # 获取列数
        for inrows in range(nrows):
            res_temp = dict()
            for incols in range(ncols):
                res_temp.update({table.cell(0, incols).value: table.cell(inrows, incols).value})
            # print(res_temp)
            res.append(res_temp)
    return res


def read_txt(text_name, path=None):
    try:
        if path is None:
            f = open("%s.txt" %text_name, encoding='utf-8')
        else:
            f = open("%s/%s.txt" % (path, text_name), encoding='utf-8')
        res = f.read()
        return res
    except:
        return


def read_file(file_name, suffix_name, path=None):
    try:
        if path is None:
            f = open("\%s.%s" % (file_name, suffix_name))
        else:
            f = open("\%s\%s.%s" % (path, file_name, suffix_name))
        res = f.read()
        return res
    except:
        return


def dir_file_list(file_dir):
    file_list = list()
    for root, dirs, files in os.walk(file_dir):
        # print(root)  # 当前目录路径
        # print(dirs)  # 当前路径下所有子目录
        # print(files)  # 当前路径下所有非目录子文件
        file_list.extend(files)
    return file_list


def get_file_size(file_dir):
    # 获取文件大小
    size_byte = os.path.getsize(file_dir)  # 字节
    size_kb = int(size_byte / 1024)
    size_mb = round(size_byte / (1024 * 1024), 2)
    size_gb = round(size_byte / (1024 * 1024 * 1024), 2)
    if size_gb < 1:
        if size_mb < 1:
            size_str = str(size_kb) + ' KB'
        else:
            size_str = str(size_mb) + ' MB'
    else:
        size_str = str(size_gb) + ' GB'
    temp_dict = {
        'size_str': size_str,
        'size_byte': size_byte
    }
    return temp_dict


class DateEncoder(json.JSONEncoder):
    # 处理json.dumps中会出现的datetime格式无法转换问题：json.dumps(each, cls=DateEncoder)
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return obj.strftime("%Y-%m-%d %H:%M:%S")
        else:
            return json.JSONEncoder.default(self, obj)


def save_list(
        file_name,
        list_data,
        split_by='\n',
        postfix="txt",
        path=None,
        overwrite=False
):
    """
    将list按照行存储到文件中，对于datetime类型自动转换为日期时间的格式
    """
    if path is None:
        file_dir = "%s.%s" % (file_name, postfix)
    else:
        lazypath.make_path(path)
        file_dir = os.path.join(path, f'{file_name}.{postfix}')

    if overwrite is True:

        write_mode = "w"  # 覆盖
    else:
        write_mode = "a"  # 追加

    f = open(file_dir, write_mode, encoding='utf-8')
    for each in list_data:

        if isinstance(each, collections.OrderedDict):
            f.write(str(json.dumps(each, cls=DateEncoder)))
        else:
            f.write(str(each))
        f.write(split_by)
    f.close()
    return file_dir


def dict_write07excel(path, table_name, sheet_name, data):
    wb = openpyxl.Workbook()
    sheet = wb.active
    sheet.title = sheet_name
    if len(data) > 0:
        row_num = 1
        col_num = 1
        for key in data[0]:
            sheet.cell(row=row_num, column=col_num, value=key)
            col_num += 1
        row_num += 1

        for each in data:
            col_num = 1
            for key in each:
                sheet.cell(row=row_num, column=col_num, value=each.get(key))
                col_num += 1
            row_num += 1

        wb.save("%s/%s.xlsx" % (path, table_name))
        showlog.info("导出数据成功！共计%s条数据" % len(data))


def save_bytes(
        bytes_content: bytes,
        file: str
):
    """
    保存bytes为文件
    """
    with open(file, 'wb') as f:
        f.write(bytes_content)
    f.close()
    return os.path.join(sys.path[0], file)


def get_file_info(file_dir):
    """
    获取文件信息
    """
    from lazysdk import lazytime
    file_path = os.path.dirname(file_dir)  # 获取文件路径
    file_name = os.path.basename(file_dir)  # 获取不含路径信息的文件名
    file_name_pure, file_suffix = os.path.splitext(file_name)  # 获取不含后缀名的文件名和后缀名
    return {
        'name': file_name,  # 获取不含路径信息的文件名
        'path': file_path,
        'pure_name': file_name_pure,
        'suffix': file_suffix,
        'size': get_file_size(file_dir=file_dir),
        'create_timestamp': os.path.getctime(file_dir),
        'create_time': lazytime.get_timestamp2datetime(os.path.getctime(file_dir)),
        'modified_timestamp': os.path.getmtime(file_dir),
        'modified_time': lazytime.get_timestamp2datetime(os.path.getmtime(file_dir)),
    }


def zip_file(
        source_file: str,
        zip_file: str = None
) -> str:
    """
    压缩文件，此方法只适用于压缩单个文件，不适合压缩文件夹
    """
    import zipfile
    file_path = os.path.dirname(source_file)
    file_name = os.path.basename(source_file)
    file_name_pure, file_suffix = os.path.splitext(file_name)
    if not zip_file:
        zip_file = os.path.join(file_path, f'{file_name_pure}.zip')
    my_zip = zipfile.ZipFile(
        file=zip_file,  # 打开一个对象，也就是目标文件名
        mode='w'  # 写入模式
    )
    my_zip.write(
        filename=source_file,  # 压缩目标文件路径
        arcname=os.path.basename(source_file),  # 内部文件名
        compress_type=zipfile.ZIP_DEFLATED
    )
    my_zip.close()
    return zip_file


def zip_path(
        source_path
) -> str:
    """
    压缩文件夹，并在原来的同级目录输出，返回zip文件路径
    """
    import shutil
    return shutil.make_archive(
        root_dir=source_path,
        base_name=source_path,
        format='zip'
    )


def make_zip(
        file: str
) -> str:
    """
    创建压缩文件，自动判断源是文件还是路径
    """
    if os.path.isdir(file):
        return zip_path(file)
    elif os.path.isfile(file):
        return zip_file(file)
    else:
        return ''


def unzip(
        file
) -> list:
    """
    解压文件，会在源文件的同级目录增加一个路径，以存放解压后的文件
    返回的是一个已解压的所有文件路径的list
    """
    import zipfile
    import locale
    file_path = os.path.dirname(file)
    file_name = os.path.basename(file)
    file_name_pure, file_suffix = os.path.splitext(file_name)
    encoding = locale.getpreferredencoding()
    my_zip = zipfile.ZipFile(
        file=file,
        mode='r'
    )
    unzip_path = os.path.join(file_path, file_name_pure)
    while True:
        if os.path.exists(unzip_path):
            unzip_path = f'{unzip_path}_unpack'
        else:
            break

    # 生成一个源文件名对照正确文件名的字典
    name_dict = dict()
    for each in my_zip.filelist:
        each_filename = copy.deepcopy(each.filename)
        if each_filename[0:8] == '__MACOSX':
            continue  # 屏蔽mac下的__MACOSX路径
        else:
            if each_filename.isascii():
                pass   # 如果isascii=True，则不改编码
            else:
                try:
                    each_filename = each_filename.encode('437').decode(encoding)  # 尝试使用437编码先解码，再编码
                except:
                    pass
            name_dict[each.filename] = each_filename

    unzip_names = list()
    for each_name in my_zip.namelist():
        new_name = name_dict.get(each_name)
        if new_name:
            new_name_dir = os.path.join(unzip_path, new_name)
            extract_name = my_zip.extract(
                member=each_name,
                path=unzip_path
            )
            if extract_name == new_name_dir:
                # print(extract_name)
                pass
            else:
                # print(extract_name, '-->', new_name_dir)
                os.renames(old=extract_name, new=new_name_dir)  # 对乱码文件名重命名
                if os.path.isfile(new_name_dir):
                    unzip_names.append(new_name_dir)
        else:
            continue
    return unzip_names


def get_stream(
        url,
        verify=True,
        **kwargs
):
    """
    以流式传输方式发起请求，适用于下载文件时，当文件过大，不会立即下载。
    当把get函数的stream参数设置成True时，它不会立即开始下载，当你使用iter_content或iter_lines遍历内容或访问内容属性时才开始下载。需要注意一点：文件没有下载之前，它也需要保持连接。
    iter_content：一块一块的遍历要下载的内容
    iter_lines：一行一行的遍历要下载的内容
    """
    try:
        response = requests.get(
            url=url,
            verify=verify,
            stream=True,
            **kwargs
        )
    except exceptions.ConnectionError:
        raise exceptions.ConnectionError(f"Could not reach host. Are you offline?")
    return response



#!/usr/bin/env python3
# coding = utf8
"""
@ Author : ZeroSeeker
@ e-mail : zeroseeker@foxmail.com
@ GitHub : https://github.com/ZeroSeeker
@ Gitee : https://gitee.com/ZeroSeeker
"""
import requests
import hashlib


def md5_file(
        file_path: str = None,
        file_url: str = None
):
    """
    计算文件的md5
    :param file_path: 本地文件路径
    :param file_url: 网络文件地址
    """
    # 文件名不影响md5
    d5 = hashlib.md5()
    if file_path:
        with open(r'%s' % file_path, 'rb') as f:
            while True:
                data = f.read(2048)
                if not data:
                    break
                d5.update(data)  # update添加时会进行计算
        return d5.hexdigest()
    elif file_url:
        resp = requests.get(file_url)
        d5.update(resp.content)  # update添加时会进行计算
        return d5.hexdigest()
    else:
        return


def md5_str(
        content,
        encoding='UTF-8'
):
    d5 = hashlib.md5()
    d5.update(content.encode(encoding=encoding))  # update添加时会进行计算
    return d5.hexdigest()



#!/usr/bin/env python3
# coding = utf8
"""
@ Author : ZeroSeeker
@ e-mail : zeroseeker@foxmail.com
@ GitHub : https://github.com/ZeroSeeker
@ Gitee : https://gitee.com/ZeroSeeker

相关文档：
https://docs.python.org/zh-cn/3/library/multiprocessing.html
"""
import random
from multiprocessing import Process
from multiprocessing import Queue
import showlog
import time
import copy
import os
os_cpu_count = os.cpu_count()  # CPU核心数


def run(
        task_list: list,
        task_function,
        subprocess_keep: bool = False,
        subprocess_limit: int = None,
        master_process_delay: int = 1,
        return_data: bool = False,
        silence: bool = True,
        task_run_time: int = None,
        task_over_time_reboot: bool = True
):
    """
    多进程 进程控制
    :param task_list: 任务列表，list格式，会将list中的每个元素传入给task_function中的task_info，作为任务的详细信息；
    :param task_function: 子任务的function，需提前写好，入参为：(task_index, task_info)，例如：task_function(task_index, task_info)
    :param subprocess_keep: 是否保持子进程，True为保持进程，死掉会自动重启；False为不保持，自然退出
    :param subprocess_limit: 进程数限制，0为无限制，否则按照设定的数量限制并行的子进程数量
    :param master_process_delay: 主进程循环延时，单位为秒，默认为1秒
    :param return_data: 是否返回数据，True返回，False不返回
    :param silence: 静默模式，为True是不产生任何提示
    :param task_run_time: 单次任务运行时长限制，单位为秒
    :param task_over_time_reboot: 单次任务超时是否重启

    demo:
    def task_function(
            task_index,
            task_info
    ):
        # 进程详细的内容
        print(task_index, task_info)
    """
    inner_task_list = copy.deepcopy(task_list)  # 深度拷贝，防止篡改
    if subprocess_limit:
        # 存在自定义的子进程数量限制，将采用
        pass
    else:
        # 不存在自定义的子进程数量限制，将使用默认计算方式
        if os_cpu_count > 1:
            # 如果cpu核心数大于1个
            subprocess_limit = os_cpu_count - 1  # 子进程数设置为cpu核心数减1
        else:
            # 如果cpu核心数等于1个
            subprocess_limit = 1
    active_process = dict()  # 存放活跃进程进程，以task_index为key，进程信息为value的dict
    total_task_num = len(inner_task_list)  # 总任务数量
    task_index_start = 0  # 用来计算启动的累计进程数
    if return_data:
        q = Queue()  # 生成一个队列对象，以实现进程通信
    else:
        pass

    if silence:
        pass
    else:
        showlog.info(f'[P-MASTER] 正在准备多进程执行任务，总任务数为：{total_task_num}，进程数限制为：{subprocess_limit}...')
    # 创建并启动进程
    while True:
        this_time_start = copy.deepcopy(task_index_start)  # 深度拷贝累积启动的进程数，以确定本次循环的起点任务序号，假设subprocess_keep为False
        for task_index in range(this_time_start, total_task_num):  # 按照任务量遍历
            # 判断是否需要创建新的子进程
            if len(active_process.keys()) >= subprocess_limit:
                # 当前活跃进程数量达到子进程数限制，本次循环不再新增进程，跳出
                if silence is False:
                    showlog.info(f'[P-MASTER] 达到子进程数限制：{subprocess_limit}')
                break
            else:
                # 未达到进程数限制
                if task_index in active_process.keys():
                    # 进程已存在，不重复创建，跳过
                    continue
                else:
                    # 进程不存在，待定
                    # 不存在子进程限制规则/当前活跃进程数量未达到进程数限制，将开启一个新进程
                    if silence is False:
                        showlog.info(f'[P-MASTER] 发现需要开启的子进程：{task_index}/{total_task_num}')
                    task_info = inner_task_list[task_index]  # 提取将开启的进程的任务内容
                    # ---------- 开启进程 ----------
                    if return_data is True:
                        p = Process(
                            target=task_function,
                            args=(task_index, task_info, q)
                        )
                    else:
                        p = Process(
                            target=task_function,
                            args=(task_index, task_info)
                        )
                    p.start()
                    # ---------- 开启进程 ----------
                    active_process[task_index] = {
                        'task_index': task_index,  # 任务序号
                        'task_info': task_info,  # 任务详情
                        'process': p,  # 进程对象
                        'task_start_time': time.time()
                    }  # 记录开启的进程
                    if silence is False:
                        showlog.info(f'[P-MASTER] 子进程：{task_index}/{total_task_num} 已开启')
                    task_index_start += 1  # 记录累计开启进程数

        # 检测非活跃进程，并从active_process中剔除非活跃进程，以便开启新的进程
        inactive_process_temp = list()  # 非活跃进程
        for process_index, process_info in active_process.items():
            # print(q.qsize())
            # print(q.get())
            # print(q.get_nowait())
            if process_info['process'].is_alive() is True:
                # 该子进程仍然运行
                if task_run_time and int(time.time() - process_info['task_start_time']) > task_run_time:
                    if not silence:
                        showlog.warning(f'[P-MASTER] 子进程：{process_info["task_index"]}/{total_task_num} 运行超时，正在关闭...')
                    process_info["process"].terminate()
                    process_info["process"].join()
                    if not silence:
                        showlog.warning(f'[P-MASTER] 子进程：{process_info["task_index"]}/{total_task_num} 运行超时，已关闭')
                    if task_over_time_reboot:
                        if not silence:
                            showlog.warning(f'[P-MASTER] 子进程：{process_info["task_index"]}/{total_task_num} 运行超时，正在重启...')
                        # ---------- 开启进程 ----------
                        if return_data is True:
                            p = Process(
                                target=task_function,
                                args=(process_info['task_index'], process_info['task_info'], q)
                            )
                        else:
                            p = Process(
                                target=task_function,
                                args=(process_info['task_index'], process_info['task_info'])
                            )
                        p.start()
                        # ---------- 开启进程 ----------
                        active_process[process_info['task_index']] = {
                            'task_index': process_info['task_index'],  # 任务序号
                            'task_info': process_info['task_info'],  # 任务详情
                            'process': p,  # 进程对象
                            'task_start_time': time.time()
                        }  # 记录开启的进程
                        if not silence:
                            showlog.warning(f'[P-MASTER] 子进程：{process_info["task_index"]}/{total_task_num} 运行超时，已重启')
                    else:
                        pass
                continue
            else:
                # 该子进程停止运行
                if silence is False:
                    showlog.info(f'[P-MASTER] 进程 {process_index}/{total_task_num} 不活跃，将被剔除...')
                inactive_process_temp.append(process_index)

        if inactive_process_temp:
            # 存在需要剔除的子进程
            for each_dead_process in inactive_process_temp:
                # 尝试终止进程
                active_process[each_dead_process]['process'].terminate()
                active_process[each_dead_process]['process'].join()
                active_process.pop(each_dead_process)
        else:
            # 不存在需要剔除的子进程
            pass
        if silence is False:
            showlog.info(f'[P-MASTER] 当前活跃进程：count:{len(active_process.keys())} --> index:{active_process.keys()}')
        else:
            pass

        if task_index_start >= len(inner_task_list) and len(active_process.keys()) == 0:
            if silence is False:
                showlog.info('[P-MASTER] 全部任务执行完成')
            else:
                pass
            if subprocess_keep is True:
                task_index_start = 0  # 将累计启动进程数重置为0
            else:
                return
        else:
            pass
        time.sleep(master_process_delay)


def run_v2(
        task_list: list,
        task_function,
        subprocess_keep: bool = False,
        subprocess_limit: int = None,
        master_process_delay: int = 1,
        # return_data: bool = False,
        silence: bool = False,
        task_run_time_limit: int = None,
        task_over_time_reboot: bool = True
):
    """
    多进程 进程控制
    在一开始就生成好全部的任务详情，按照任务详情去具体执行和控制

    :param task_list: 任务列表，list格式，会将list中的每个元素传入给task_function中的task_info，作为任务的详细信息；
    :param task_function: 子任务的function，需提前写好，入参为：(task_index, task_info)，例如：task_function(task_index, task_info)
    :param subprocess_keep: 是否保持子进程，True为保持进程，死掉会自动重启；False为不保持，自然退出
    :param subprocess_limit: 进程数限制，0为无限制，否则按照设定的数量限制并行的子进程数量
    :param master_process_delay: 主进程循环延时，单位为秒，默认为1秒
    # :param return_data: 是否返回数据，True返回，False不返回
    :param silence: 静默模式，为True是不产生任何提示
    :param task_run_time_limit: 单次任务运行时长限制，单位为秒
    :param task_over_time_reboot: 单次任务超时是否重启

    demo:
    def task_function(
            task_index,
            task_info
    ):
        # 进程详细的内容
        print(task_index, task_info)
    """
    # process_task_list = copy.deepcopy(task_list)  # 深度拷贝，防止篡改源数据
    if task_list:
        pass
    else:
        if not silence:
            showlog.warning('无任务需要执行')
        return

    active_process = dict()  # 存放活跃进程进程，以task_index为key，进程信息为value的dict

    # 任务初始化
    process_task_dict = dict()
    for each_index, each_task in enumerate(task_list):
        task_index = copy.deepcopy(each_index+1)
        process_task_dict[task_index] = {
            'task_index': task_index,
            'task_info': copy.deepcopy(each_task),
            'task_state': 0  # 默认为等待状态
        }
    total_task_num = len(process_task_dict.keys())  # 总任务数量

    # ---------------------- 决定子进程数量限制 ----------------------
    if subprocess_limit:
        # 存在自定义的子进程数量限制，将采用
        pass
    else:
        # 不存在自定义的子进程数量限制，将使用默认计算方式
        if os_cpu_count > 1:
            # 如果cpu核心数大于1个
            subprocess_limit = os_cpu_count - 1  # 子进程数设置为cpu核心数减1
        else:
            # 如果cpu核心数等于1个
            subprocess_limit = 1
    # ---------------------- 决定子进程数量限制 ----------------------

    if not silence:
        showlog.info(f'[P-MASTER] 正在准备多进程执行任务，总任务数为：{total_task_num}，进程数限制为：{subprocess_limit}...')
    # 创建并启动进程
    proces_running_count = 0  # 用来计算正在运行的进程数
    while True:
        finish_count = 0
        for task_index, task_info in process_task_dict.items():
            process = task_info.get('process')
            task_start_time = task_info.get('task_start_time')
            task_state = task_info.get('task_state')

            if task_state == 0:  # 任务处于待开启状态
                # 检查进程数限制
                if not subprocess_limit:
                    # 无进程数限制，直接将任务打开
                    task_info['task_state'] = 1
                elif subprocess_limit and proces_running_count < subprocess_limit:
                    # 有进程数限制，但是未达到限制，打开任务
                    task_info['task_state'] = 1
                else:
                    # 有进程数限制，且达到了限制，什么也不做
                    pass
            elif task_state == 1:  # 任务处于开启状态
                pass
            else:
                finish_count += 1
                continue  # 当任务状态不属于等待和待开启状态时，表示该进程结束，将跳过检查

            # 先检查进程是否存活
            if process:
                if process.is_alive():
                    # 进程存在且进程存活
                    if task_run_time_limit:
                        # 存在进程最大运行时间限制
                        if int(time.time() - task_start_time) > task_run_time_limit:
                            # 运行超时
                            if not silence:
                                showlog.warning(f'[P-MASTER] 子进程：{task_index}/{total_task_num} 运行超时，正在关闭...')
                            process.terminate()
                            process.join()
                            proces_running_count -= 1
                            if not silence:
                                showlog.warning(f'[P-MASTER] 子进程：{task_index}/{total_task_num} 运行超时，已关闭')

                            if task_over_time_reboot:
                                # 超时将重启
                                if not silence:
                                    showlog.warning(f'[P-MASTER] 子进程：{task_index}/{total_task_num} 运行超时，将重启...')
                                task_info.pop('process')
                            else:
                                # 超时不重启，直接标记为结束
                                task_info['task_state'] = 2
                        else:
                            # 运行未超时
                            pass
                    else:
                        # 不存在进程最大运行时间限制
                        pass
                else:
                    # 进程存在但已结束
                    if not silence:
                        showlog.info(f'[P-MASTER] 进程 {task_index}/{total_task_num} 已结束 <dead>')
                    process.terminate()
                    process.join()
                    task_info['task_state'] = 2
                    proces_running_count -= 1
            else:
                # 进程不存在，需要判断是否开启进程
                pass

            # 开启进程
            if task_info['task_state'] == 1 and task_info.get('process') is None:
                if silence is False:
                    showlog.info(f'[P-MASTER] 发现需要开启的子进程：{task_index}/{total_task_num}')
                p = Process(
                    target=task_function,
                    args=(task_index, task_info)
                )
                p.start()
                task_info['process'] = p
                task_info['task_start_time'] = time.time()
                proces_running_count += 1
            else:
                continue
        if not subprocess_keep and finish_count == total_task_num:
            break
        else:
            time.sleep(master_process_delay)


def task_function_demo(
        task_index,
        task_info,
        q=None
):
    showlog.info(f'[p-{task_index}] start')
    # print(task_index, task_info)
    count_down = task_info.get('count_down')
    this_randint = random.randint(0,10)
    print(task_index, 'this_randint:', this_randint)

    time.sleep(this_randint)
    # q.put(task_index)
    showlog.info(f'[p-{task_index}] finish')


if __name__ == '__main__':
    """
    尝试增加对进程运行时间计时，超时的将关闭重启
    """
    task_list_demo = [
        {'task_id': 1, 'count_down': 1},
        {'task_id': 2, 'count_down': 2},
        {'task_id': 3, 'count_down': 3},
        {'task_id': 4, 'count_down': 4},
        {'task_id': 5, 'count_down': 5},
        {'task_id': 6, 'count_down': 6},
        {'task_id': 7, 'count_down': 7},
        {'task_id': 8, 'count_down': 8},
        {'task_id': 9, 'count_down': 9},
        {'task_id': 10, 'count_down': 10},
        {'task_id': 11, 'count_down': 11},
    ]
    run_v2(
        task_list=task_list_demo,
        task_function=task_function_demo,
        task_run_time_limit=5
        # return_data=True,
        # silence=True
    )



import urllib.parse as urlparse


def get_url_params(
        url: str
):
    """
    获取url的params参数，返回dict形式
    """
    params_str = urlparse.urlsplit(url).query
    if params_str:
        params_str_split = params_str.split('&')
        params_dict = dict()
        for each in params_str_split:
            each_split = each.split('=', maxsplit=1)
            params_dict[each_split[0]] = each_split[1]
        return params_dict
    else:
        return


def url_info(
        url: str
):
    url_info_dict = dict()
    url_info_dict['url'] = url
    if url:
        urlparse_obj = urlparse.urlsplit(url)
        url_info_dict['host'] = urlparse_obj.hostname  # 域名
        url_info_dict['path'] = urlparse_obj.path  # 路径
        url_info_dict['scheme'] = urlparse_obj.scheme  # 协议
        url_info_dict['params'] = get_url_params(url)
    else:
        pass
    return url_info_dict


def url_quote(url):
    """
    url编码
    """
    return urlparse.quote(url)


def url_unquote(url):
    """
    url解码
    """
    return urlparse.unquote(url)


def domain_split(url):
    """
    对输入的url分析域名，拆分域名
    domain_1: 顶级域名
    domain_2: 二级域名
    """
    import urllib
    if 'http' in url:
        netloc = urllib.parse.urlparse(url).netloc
    else:
        netloc = url
    netloc_split = netloc.split('.')
    domain_dict = dict()
    for _ in range(len(netloc_split)):
        domain_dict[f'domain_{_+1}'] = '.'.join(netloc_split[-_ - 1:])
    return domain_dict
